"""Good uses of range."""

import six

if six.PY3:
    def xrange(*args_unused):       # pylint: disable=unused-argument, redefined-builtin
        """Just to keep PY3 happy."""


i = 2

range(10)
range(i)
range(1, 10)
range(1, 10, 2)
range(i, 10)
range(0, 10, 2)
xrange(10)
xrange(i)
xrange(1, 10)
xrange(1, 10, 2)
xrange(i, 10)
xrange(0, 10, 2)



"""A TestCase class that doesn't call super setUp or super tearDown."""

import unittest

# pylint: disable=too-few-public-methods, missing-docstring

class BadTestCase(unittest.TestCase):
    def setUp(self):
        self.i_am_bad = True

    def tearDown(self):
        self.i_am_bad = True

class AllowedToBeHalfBadTestCase(unittest.TestCase):
    # pylint: disable=super-method-not-called
    def setUp(self):
        self.i_am_bad = True
    # pylint: enable=super-method-not-called

    def tearDown(self):
        self.i_am_bad = True



"""Test main for edx-lint."""

import os
import unittest
import re

def load_tests(unused_loader, tests, unused_pattern):  # pylint: disable=unused-argument
    """Loads tests for the pylint test loader.

    This function is automatically run by pylint's test runner, and is called
    with three arguments, two of which we don't need.

    """
    # Have to import this in the function, because the module does
    # initialization on import! ugh.
    from pylint.testutils import LintTestUsingFile, linter, get_tests_info

    # Inline functions that was deleted from pylint core
    def cb_test_gen(base_class):
        """Inlined from pylint"""
        def call(input_dir, msg_dir, module_file, messages_file, dependencies):
            """Inlined from pylint"""
            class LintTC(base_class):
                """Inlined from pylint"""
                module = module_file.replace('.py', '')
                output = messages_file
                depends = dependencies or None
                INPUT_DIR = input_dir
                MSG_DIR = msg_dir
            return LintTC
        return call

    def make_tests(input_dir, msg_dir, filter_rgx, callbacks):
        """generate tests classes from test info

        return the list of generated test classes
        """
        if filter_rgx:
            is_to_run = re.compile(filter_rgx).search
        else:
            is_to_run = lambda x: 1
        tests = []
        for module_file, messages_file in (
                get_tests_info(input_dir, msg_dir, 'func_', '')
        ):
            if not is_to_run(module_file) or module_file.endswith(('.pyc', "$py.class")):
                continue
            base = module_file.replace('func_', '').replace('.py', '')

            dependencies = get_tests_info(input_dir, msg_dir, base, '.py')

            for callback in callbacks:
                test = callback(input_dir, msg_dir, module_file, messages_file,
                                dependencies)
                if test:
                    tests.append(test)
        return tests


    # Load our plugin.
    linter.load_plugin_modules(['edx_lint.pylint'])

    here = os.path.dirname(os.path.abspath(__file__))

    tests = make_tests(
        input_dir=os.path.join(here, 'input'),
        msg_dir=os.path.join(here, 'messages'),
        filter_rgx=None,
        callbacks=[cb_test_gen(LintTestUsingFile)],
    )

    cls = unittest.TestSuite
    return cls(unittest.makeSuite(test, suiteClass=cls) for test in tests)



"""Unit tests for required-base-classes."""

import astroid
from pylint.testutils import CheckerTestCase, Message, set_config

from edx_lint.pylint.required_base_class import RequiredBaseClassChecker


class RequiredBaseClassTestCase(CheckerTestCase):
    """Unittest tests of RequiredBaseClassChecker."""

    CHECKER_CLASS = RequiredBaseClassChecker

    def test_no_messages_by_default(self):
        node = astroid.parse('''
            class MyClass(object):
                pass
        ''')
        with self.assertNoMessages():
            self.walk(node)

    @set_config(required_base_class=["BaseClass:MyMixin"])
    def test_no_messages_if_class_not_used(self):
        node = astroid.parse('''
            class MyClass(object):
                pass
        ''')
        with self.assertNoMessages():
            self.walk(node)

    @set_config(required_base_class=["unittest.case.TestCase:.MyTestMixin"])
    def test_error_if_class_is_not_used(self):
        node = astroid.parse('''
            from unittest import TestCase
            class MyClass(TestCase):
                pass
        ''')
        expected_msg = Message(
            'missing-required-base-class',
            node=node.body[-1],
            args=('MyClass', '.MyTestMixin'),
        )
        with self.assertAddsMessages(expected_msg):
            self.walk(node)

    @set_config(required_base_class=["unittest.case.TestCase:.MyTestMixin"])
    def test_no_messages_if_class_is_used(self):
        node = astroid.parse('''
            from unittest import TestCase
            class MyTestMixin(object):
                pass
            class MyClass(MyTestMixin, TestCase):
                pass
        ''')
        with self.assertNoMessages():
            self.walk(node)

    @set_config(required_base_class=["unittest.case.TestCase:.MyTestMixin"])
    def test_old_style_classes(self):
        # We don't support base class checking on old-style classes, but we
        # have to be sure not to fall over at least.
        node = astroid.parse('''
            class MyClass:
                pass
        ''')
        with self.assertNoMessages():
            self.walk(node)



"""Pylint plugin: some classes have required base classes."""

import collections

from pylint.checkers import BaseChecker, utils
from pylint.interfaces import IAstroidChecker

from .common import BASE_ID


def register_checkers(linter):
    """Register checkers."""
    linter.register_checker(RequiredBaseClassChecker(linter))


class RequiredBaseClassChecker(BaseChecker):
    """Pylint checker for required base classes."""

    __implements__ = (IAstroidChecker,)

    name = 'required-base-class-checker'

    MESSAGE_ID = "missing-required-base-class"
    msgs = {
        'E%d40' % BASE_ID: (
            "class %s is missing required base class %s",
            MESSAGE_ID,
            "Used when a class is missing a base class required by one of its "
            "other base classes.",
        ),
    }

    options = [
        ('required-base-class', {
            'default' : (),
            'type': 'csv',
            'metavar': '<class name pairs>',
            'help': 'Fooey'}
        ),
    ]

    def __init__(self, linter):
        super(RequiredBaseClassChecker, self).__init__(linter)
        self.class_map = collections.defaultdict(set)

    def open(self):
        # pylint: disable=no-member
        if self.config.required_base_class:
            for pair in self.config.required_base_class:
                child, parent = pair.split(":")
                self.class_map[child].add(parent)

    @utils.check_messages(MESSAGE_ID)
    def visit_class(self, node):
        """Check each class."""
        if not self.class_map:
            return

        try:
            all_bases = [usable_class_name(c) for c in node.mro()]
        except NotImplementedError:
            # Old-style 2.x classes have no mro.
            all_bases = []

        for base in all_bases:
            required = self.class_map.get(base)
            if required is not None:
                if not all(r in all_bases for r in required):
                    nice_required = ", ".join(sorted(required))
                    self.add_message(self.MESSAGE_ID, args=(node.name, nice_required), node=node)


def usable_class_name(node):
    """Make a reasonable class name for a class node."""
    name = node.qname()
    for prefix in ["__builtin__.", "builtins."]:
        if name.startswith(prefix):
            name = name[len(prefix):]
    return name
